(require [kodhy.macros [ss ssi $ λ qw getl geti wc cbind ordf
  rmap amap replicate geta filt afind ecase]])

(import
  [hy.contrib.hy-repr [hy-repr]]
  os re
  [collections [OrderedDict]]
  [numpy :as np]
  [scipy.stats :as scist]
  [pandas :as pd]
  [matplotlib.pyplot :as plt]
  [seaborn :as sns]
  sklearn.linear-model
  sklearn.model-selection
  [statsmodels.stats.proportion [proportion-confint]]
  scikits.bootstrap
  [kodhy.util [T F seq rd pretty-json-to-pd valcounts dotplot rectplot
    cat cols2map concat ucfirst
    pairs pds-from-pairs drop-unused-cats R-call cols2map density-plot
    cached-eval]])

(setv mean-days-in-month 30.436875)
(setv mean-weeks-in-month (/ mean-days-in-month 7))
(setv drinks-amount-map (OrderedDict (pairs
  "1 or 2"      1.5
  "3 or 4"      3.5
  "5 or 6"      5.5
  "7,8 or 9"    8
  "10 or more" 10)))
(setv drinks-freq-map {
  "drinks/month" (OrderedDict (pairs
    "Less than once a month" 0.5
    "Once a month"           1
    "2 to 3 times a month"   2.5
    "Once a week"            mean-weeks-in-month
    "2 times a week"         (* 2 mean-weeks-in-month)
    "3 to 4 times a week"    (* 3.5 mean-weeks-in-month)
    "Nearly every day"       (* .75  mean-days-in-month)
    "Every day"              mean-days-in-month))
  "aa/day" (OrderedDict (pairs
    ; Mary O'Connor's suggestions
    "Less than once a month"  .02
    "Once a month"            .05
    "2 to 3 times a month"    .1
    "Once a week"             .2
    "2 times a week"          .5
    "3 to 4 times a week"     .6
    "Nearly every day"        .7
    "Every day"              1.0))})

(setv obs (pretty-json-to-pd "../data.json"))

(setv s-late-entry (frozenset (.difference (set ($ obs s))
  (ss obs (= $timepoint "Baseline") "s"))))

(setv conditions (cols2map obs "s" "condition" :keep "first"))
(setv nhood-conditions (cols2map obs "nhood" "condition" :keep "first"))

(when (in "interview_date" obs.columns)
  (for [datecol (qw interview_date child_birthdate)]
    (setv (getl obs : datecol) (pd.to-datetime (getl obs : datecol)
      :format "%Y-%m-%d" :exact T)))
  (setv mean-day-numbers ((fn []
    (setv p (.pivot obs "s" "timepoint" "interview_date"))
    (dict (rmap [timepoint p.columns]
      (setv v (.dropna (- (getl p : timepoint) (getl p : "Baseline"))))
      (, timepoint (.mean v.dt.days)))))))
  (setv ($ obs child_age_days) (wc obs
    (. (- $interview_date $child_birthdate) dt days))))

(when (in "alcohol_problem" obs.columns)
  (setv ($ obs alcohol_state) (wc obs
    (np.where (pd.isnull $alcohol_any) None
    (np.where (= $alcohol_any "Sober") "Sober"
                                       $alcohol_problem))))
  (setv ($ obs alcohol_state) (.astype ($ obs alcohol_state) 
    (pd.api.types.CategoricalDtype
      :ordered T :categories (qw Sober Okay Problem))))
  (setv ($ obs prenatal_alcohol_state) (.map ($ obs s)
    (cols2map (ss obs (= $timepoint "Baseline")) "s" "alcohol_state"))))

; Clip the growth z-scores to sensible bounds.
(for [vname (qw child_height_for_age_z child_weight_for_age_z)]
  (when (in vname obs.columns)
    (setv (getl obs : vname) (np.clip (getl obs : vname) -5 5))))

(setv ($ obs alcohol_any_preg_before_noticed) (pd.Categorical.from-codes
  :categories ["Sober" "Drinking"]
  :codes (rmap [v ($ obs alcohol_freq_preg_before_noticed)]
    (cond
      [(pd.isnull v) -1]
      [(= v "Never") 0]
      [T 1]))))

(defn alc-freq-code [drink-multiplier freq-map] (wc obs
  (np.where (pd.isnull $alcohol_freq_preg_before_noticed) np.nan
  (np.where (= $alcohol_freq_preg_before_noticed "Never") 0
  (np.where (pd.isnull $alcohol_amount_preg_before_noticed) np.nan
  (*
    (.map $alcohol_freq_preg_before_noticed freq-map)
    drink-multiplier
    (.map $alcohol_amount_preg_before_noticed drinks-amount-map)))))))

(setv ($ obs alcohol_drinks/month_preg_before_noticed)
  (alc-freq-code
    :drink-multiplier 1
    :freq-map (get drinks-freq-map "drinks/month")))

(setv ($ obs alcohol_aa/day_preg_before_noticed)
  (alc-freq-code
    :drink-multiplier .6 ; A standard drink has .6 fl oz of pure alcohol
    :freq-map (get drinks-freq-map "aa/day")))

(defn countbars [varname &optional timepoint]
  (setv v (.value-counts (if timepoint
    (ss obs (= $timepoint timepoint) varname)
    (getl obs : v))))
  (plt.bar v.index v :width 1))

(defn cache [key f &optional bypass]
  (cached-eval key f bypass (os.path.join (os.path.expanduser "~")
     ".daylight" "py-cache" "Kodi-projects" "philani")))

(setv alc-terms-d (dict
  :no_alc ""
  :prenatal "+ prenatal_alcohol_state + prenatal_alcohol_state:condition"
  :current "+ alcohol_state + alcohol_state:condition"))
(setv (get alc-terms-d "both")
  (+ (get alc-terms-d "prenatal") (get alc-terms-d "current")))

(setv binomial-yvar-trials (dict
  :child_execufun_silly_sounds 36
  :child_execufun_operation_span 20
  :child_execufun_something_same 36))

(defn alc-model-aics [&rest args &kwargs kwargs]
  (cache
    (hy-repr (+ (, 'alc-model-aics) args
      (if kwargs (tuple (.items kwargs)) (,))))
    (fn [] (dfor
      k (.keys alc-terms-d)
      [k (alc-model-aic k #* args #** kwargs)]))))
(defn show-alc-model-aics [aics]
  (setv [ks vs] (zip #* (.items aics)))
  (.round (pd.Series
    :index (+ (, "n_obs" "n_subjects" "n_timepoints") ks)
    (+ [(get vs 0 "n_obs") (get vs 0 "n_subjects") (get vs 0 "n_timepoints")]
      (amap (get it "aic") vs)))))

(defn alc-model-aic [alc-terms yvar yvar-type &optional clip]
  (setv [d ivs model-f] (alc-model-prep alc-terms yvar yvar-type clip))
  {
    "n_obs" (first d.shape)
    "n_subjects" (len (.unique ($ d s)))
    "n_timepoints" (len (.unique ($ d timepoint)))
    "aic" (R-call
      (.format "function(d) AIC({})"
        (model-f "d" yvar ivs :reml "FALSE"))
      d)})

(defn alc-model-coefcis [alc-terms yvar yvar-type &optional clip]
  (setv [d ivs model-f] (alc-model-prep alc-terms yvar yvar-type clip))
  (setv coefs (cache
    (hy-repr ['alc-model-coefcis-R-call alc-terms yvar yvar-type])
    (fn [] (->> d (R-call
      (+ "function(d)
         {m = " (model-f "d" yvar ivs :reml "TRUE") "
          if (is(m, 'merMod'))
              library(lme4)
          coefs = (if (is(m, 'merMod')) fixef else coef)(m)
          cis = (if (is(m, 'merMod'))
            confint(m, names(coefs), .95, method = 'boot') else
            confint(m))
          p = coef(summary(m))[,ncol(coef(summary(m)))]
          data.frame(name = names(coefs), coefs, cis[,1], cis[,2], p)}"))))))
  (prettify-coefcis (.set-index coefs "name")))

(defn prettify-coefcis [d]
  (setv d (.copy d))
  (setv d.columns (qw Point Low High p))
  (setv d.index (rmap [i d.index]
    (when (instance? bytes i)
      (setv i (.decode i "ASCII")))
    (re.sub r"(.+):conditionMentor_Mother" r"conditionMentor_Mother:\1" i)))
  (setv d.index (rmap [i d.index] (-> i
    (.replace "timepoint" "")
;    (.replace "prenatal_alcohol_state" "Prenatal ")
    (.replace "alcohol_state" "Current")
    (.replace "Okay" "occasional drinking")
    (.replace "Problem" "problem drinking")
    (.replace "Mentor_Mother" "Intervention")
    (.replace "condition" "")
    (.replace ":" " × ")
    (.replace "_" " "))))
  (setv d.index.name "Coefficient")
  d)

(defn coef-formatter [row]
  (setv [point lo hi p] row)
  (.format "{:.02f} [{:.02f}, {:.02f}], p {}"
    point lo hi
    (if (< p .01) "< .01" (+ "= " (.lstrip (format p ".02f") "0")))))

(defn coefcis-multi-dvs [&kwargs coefci-ds]
  (setv d (pd.concat :sort T :axis 1 (rmap [d (.values coefci-ds)]
    (.apply d :axis 1 coef-formatter))))
  (setv d.columns (.keys coefci-ds))
  (setv d.index (amap (+ "     " it) d.index))
  (setv (getl d "Main effects") NaN)
  (setv (getl d "Interactions") NaN)
  (getl d (sorted d.index :key (λ (,
    (cond
      [(= it "Main effects") 0]
      [(and (not-in "×" it) (in " " it)) 1]
      [(= it "Interactions") 2]
      [True 3])
    (not-in "(Intercept)" it)
    (not (or (in "month" it) (in "year" it)))
    (in "Intervention" it)
    (and (in "drinking" it) (in "current" (.lower it)))
    (not (in "6 months" it)))))))

(defn alc-model-prep [alc-terms yvar yvar-type clip]
  (setv d (.dropna (drop-unused-cats
    ; Drop the after-birth timepoint, since there's basically no
    ; variation of the alcohol variables there.
    (getl (ss obs (!= $timepoint "After_birth")) : (+ [yvar] (qw
      s condition timepoint alcohol_state prenatal_alcohol_state))))))
  (when clip
    (setv (getl d : yvar) (.clip (getl d : yvar) #** clip)))
  (setv mixed-model? (> (len (.unique ($ d timepoint))) 1))
  (setv base-ivs (if mixed-model?
    "0 + timepoint + timepoint:condition"
    "condition"))
  (setv ivs (+ base-ivs " " (get alc-terms-d alc-terms)))
  (setv model-f (fn [d yvar ivs reml] (cond
    [(and mixed-model? (= yvar-type "normal"))
      (.format "lmerTest::lmer({} ~ {} + (1|s), data = {}, REML = {})"
        yvar ivs d reml)]
    [(and (not mixed-model?) (= yvar-type "normal"))
      (.format "lm({} ~ {}, data = {})"
        yvar ivs d)]
    [(and mixed-model? (= yvar-type "binomial"))
      (.format "lme4::glmer(cbind({}, {} - {}) ~ {} + (1|s), data = {},
                family = binomial(link = 'logit'),
                control = lme4::glmerControl(optimizer = 'bobyqa',
                    optCtrl = list(maxfun = 2e5)))"
            yvar (get binomial-yvar-trials yvar) yvar ivs d)]
    [T
      (raise (ValueError "not implemented"))])))
  [d ivs model-f])

(defn alc-outcome-plot [yvar]
  (setv d (pd.melt
    (.dropna (getl obs : [yvar "prenatal_alcohol_state" "alcohol_state"]))
    [yvar]
    (qw prenatal_alcohol_state alcohol_state)))
  (setv ($ d value) (.astype ($ d value) "category" :categories ["Sober" "Okay" "Problem"]))
  (sns.factorplot :data d "value" yvar :col "variable"
    :kind "box" :color "#aaaaff"))

(defn coef-plot [alc-terms yvar yvar-type ylab yrange]

  (setv d1 (alc-model-coefcis alc-terms yvar yvar-type))
  (setv mdn (dfor
    [k v] (.items mean-day-numbers)
    :if (any (amap (in (.replace k "_" " ") it) d1.index))
    [(.replace k "_" " ") v]))

  (setv alctimes (+ [None]
    (if (in alc-terms (qw both prenatal)) ["Prenatal"] [])
    (if (in alc-terms (qw both current)) ["Current"] [])))
  (when (= alc-terms "both")
    ; Move the no-drinking timepoint to the end.
    (.remove alctimes None)
    (.append alctimes None))
  (setv legend-items {})

  (for [[pi alctime] (enumerate alctimes)]

    (setv d (getl d1 (filt
      (and
        (or (= alctime "Prenatal") (not-in "Prenatal" it))
        (or (= alctime "Current") (not-in "Current" it)))
      d1.index)))
    (setv x (.reset-index (getl d (.keys mdn) ["Point"])))
    (setv x.columns ["timepoint" "Control"])
    (setv ($ x Intervention) (+ ($ x Control)
      (.reset-index :drop T (getl d (amap (+ "Intervention × " it) (.keys mdn)) "Point"))))
    (when alctime
       (for [alclevel (qw problem occasional)  condition (qw Control Intervention)]
          (setv (getl x : (cat condition (and (= alclevel "problem") ", problem"))) (+
            (getl x : condition)
            (getl d (afind (in alclevel it) d.index) "Point")
            (if (= condition "Control") 0
              (getl d (afind (and (in "Intervention" it) (in alclevel it)) d.index) "Point"))))))

    (plt.subplot 1 (len alctimes) (inc pi))
    (plt.axhline :color "black" :linewidth .5)
    (for [col (filt (!= it "timepoint") x.columns)]
      (setv (get legend-items col) (first (plt.plot :marker "o"
        (.map ($ x timepoint) mdn)
        (getl x : col)
        :label col
        :color (if (in "Interven" col) "#ff8000" "#0000a0")
        :linestyle (if (in "problem" col) "dotted" "solid")))))
    (if pi
      (.yaxis.set-ticklabels (plt.gca) [])
      (plt.ylabel ylab))
    (plt.ylim (get yrange 0) (get yrange 1))
    (plt.yticks (np.arange
      (get yrange 0) (+ (get yrange 1) (get yrange 2)) (get yrange 2)))
    (plt.xlabel "Timepoint")
    (plt.xticks :rotation "vertical"
      (tuple (.values mdn))
      (.keys mdn))
    (.set-title (plt.gca) (ecase alctime
      ["Current" "Drinking currently"]
      ["Prenatal" "Drinking prenatally"]
      [None (ecase alc-terms
        ["prenatal" "Not drinking prenatally"]
        ["current" "Not drinking currently"]
        ["both" "Neither"])])))

  (plt.legend :handles (list (.values legend-items))))
